  /************************************************************
 Universal 8bit Graphics Library (https://github.com/olikraus/u8g2/)
  Copyright (c) 2016, olikraus@gmail.com
  All rights reserved.
  
  Polysynth project. Started with test code on breadboard, then real HW, then multiple voices. 
  Teensy 3.5, Clock 120 or 144MHz overclock, USB: Serial 
  D. Erickson
  
  10/27/18 Get DAC MCP4822 working with BitBang, Midi In and Out working, Scaling on DAC is excellent.
           Pitch wheel working
  11/3/18  Fast bit bang SPI, 7MHz vs 1MHz. DAC108. 
  2/18/19  Begin polysynth with Teensy 3.5, PCboards, real SPI, and polyphony. 
  3/2/19   Updated SPI and DAC. Back to bit-bang using hi-level 12b and 10b Dacs working
  3/3/19   Dig Pots working
  3/9/19   Patch structure working with mixer. Set from midi
           Coarse and fine detunes
  3/20/19  Hardware struccture
  3/31/09  PolySynth 3: Start of Polyphony
  4/6/19   Unison working
  4/7/19   Chords working, but not key off
  4/8/19   Poly mode working: basic key off. No smart voice allocation
  9/1/19   Filter tracks keyboard
*/

#include <MIDI.h>
#include <Bounce.h>
#include <U8g2lib.h>
//#include "midiCodes.h"  //Not used in PolySynth
#include <TimerOne.h>

//MIDI_CREATE_INSTANCE(HardwareSerial, Serial1, MIDI);
MIDI_CREATE_DEFAULT_INSTANCE();

const int channel = 1; // Midi Channel
int instrument = 0;

int gate1State = 0;
int average = 16;
int printnotes = 1;
int playnote;
int pitchDac = 0;
char majorScale[] = {0, 2, 4, 5, 7, 9, 11, 12};
volatile int V1dac = 0, V2dac = 0; 
int patchSel = 0;
int fineDetuneV2, coarseDetuneV2 = 0;
int  j = 0;
#define voices 3
unsigned char voice = 0;
unsigned char nextvoice = 0;      // basic voice counter 
unsigned char unison = 0;         // unison mode 


// 10b DAC Channels
#define D_FILT_F 0
#define D_FILT_R 1  
#define D_VCA 2
#define D_MOD3 3
#define D_MOD4 4
#define D_SPARE 5
#define D_PW1 6
#define D_PW2 7

// Digital pot (mux) channels
#define M_V1_TRI 0
#define M_V1_SAW 1
#define M_V1_SQU 2
#define M_NOISE  3
#define M_V2_TRI 4
#define M_V2_SAW 5
#define M_V2_SQU 6

// Envelope generator params
#define ENV_IDL 0
#define ENV_ATT 1
#define ENV_DEC 2
#define ENV_REL 3
#define ENV_MAX 10000    // Not sure best scaling here

// Interrupt (env. and LFO) sample freq. in Hz
#define F_SAMPLE 1000 

// Ports and bits for fast bit-bang SPI, Teensy 3.5
#define P_SPI_SS   PORTB     // Port B10  Pin 31
#define B_SPI_SS   10
#define P_SPI_MOSI PORTD     // Port D6
#define B_SPI_MOSI 4
#define P_SPI_SCK  PORTB     // Port B10
#define B_SPI_SCK  11       

// Oxygen 49 keyboard MIDI: Drawbars
#define OXC1  20
#define OXC2  21
#define OXC3  71
#define OXC4  72
#define OXC5  25
#define OXC6  73
#define OXC7  74
#define OXC8  70
#define OXC9  63
// Knobs
#define OXC10  22
#define OXC11  23
#define OXC12  61
#define OXC13  24
#define OXC14  26
#define OXC15  27
#define OXC16  62
#define OXC17  95

// Pin definitions
// Don't use LED pin, it provides OLED SPI clock
const int SPI_CK   = 32;       // Pin defs for digitalWrite() slow SPI
const int SPI_MOSI = 6;
const int SPI_SS   = 31;
#define BA0 26
#define BA1 25
#define BA2 24
#define CA0 35
#define CA1 36
#define CA2 37

// U8g2 Contructor List (Frame Buffer: F, U8G2_Rx : rotation)
// The complete list is here: https://github.com/olikraus/u8g2/wiki/u8g2setupcpp
// Update pin numbers according to your setup. Use U8X8_PIN_NONE if no reset pin connected
// Note Pin 13 (11) is MOSI / DO
U8G2_SH1106_128X64_NONAME_F_4W_HW_SPI u8g2(U8G2_R2, /*cs*/ 10, /*dc*/ 9, /*res*/ 8);

struct vcoSet{
 int newnote;
 char order; // 0 is free, 1: first, 2: sec...
 int V1dac;
 int V2dac;
 int pitchDac;
//  V1dac = newnote * 500 / 12 + pitchDac + cal.V1Off;  
//  V2dac = newnote * 500 / 12 + pitchDac + cal.V2Off + fineDetuneV2 + coarseDetuneV2;  
} vco[voices];

struct env {
  int att;      
  int dec;
  int sus;
  int rel;
  int level;
  int state[voices];
  int out[voices];
  int dac[voices];
} filt_env, amp_env;

struct filter {
  int freq;
  int res;
  int sel;
  int track;      // Keyboard tracking 0/1? 
} fil;

struct lfo {    // Basic LFO
  float freq;
  int wave;
  int amp;
  int ph_inc;
  int ph_acc;
  int out;

} lfo1, lfo2;

struct mixer {  // levels are 0 to 127, just multiply 2x for the hardware
  char flag;    // 1 if mixer needs updating
  int v1Tri;
  int v1Saw;
  int v1Squ;
  int noise;
  int v2Tri;
  int v2Saw;
  int v2Squ;
} mix;

struct calibration{
  float v1Offs;
  float v2Offs;
};

calibration boardCal[] = {    // Assumes boards are installed in order
  {0.0, 0.0},     // No board 0
  {1.3, 4.8},     // Board 1
  {2.2, 0.7},     // Board 2
  {0.8, 0.0},     // Board 3
};
  
// No cal
//#define V1OffSemi +0.0   
//#define V2OffSemi +0.0
// Board 1
//#define V1OffSemi +1.2    // 330K VCO Cal factors in semi-tones
//#define V2OffSemi +5.0
// Board 2
//#define V1OffSemi  2.6   // 301K 
//#define V2OffSemi  1.0
// Board 3
//#define V1OffSemi 1.8      // 301K
//#define V2OffSemi 0.7

#define V1TRI 0
#define V1SAW 1
#define V1SQU 2
#define V2TRI 3
#define V2SAW 4
#define V2SQU 5
#define NOISE 6

struct patchdata {        // Data structure for patches. These are the user and midi controls 
  char Pname[20];
  char V1Semi;
  int  V1Cent;
  char V2Semi;
  int  V2Cent;
  char Mix[7];      // TRI,SAW,SQU, TRI,SAW,SQU, noise
  int  VCAEnv[4];     // ADSR
  char FilCut;
  char FilRes;
  char FilType;
  int  FilEnv[4];
  char FilTrack;
  char Mod[8]; 
};

#define NUMPATCH 11
struct patchdata patch[] = {
//  Pname      V1  V2   Mix     V2Mix    No   VCAEnv              FC FR FTy FilEnv FTrk Mod
  { "2x Saw",  0,0,0,0,{0,127,0, 0,127,0, 0},{20,10,ENV_MAX/2,10},43,69,0, {0,0,0,0},0,{0,0,0,0,0,0,0,0}},
  { "2x Squ",  0,0,0,0,{0,0,127, 0,0,127, 0},{20,10,ENV_MAX/2,10},43,69,0, {0,0,0,0},0,{0,0,0,0,0,0,0,0}},
  { "2x Tri",  0,0,0,0,{127,0,0, 127,0,0, 0},{20,10,ENV_MAX/2,10},43,69,0, {0,0,0,0},0,{0,0,0,0,0,0,0,0}},  
  { "V1 Saw",  0,0,0,0,{0,127,0, 0,0,0,   0},{20,10,ENV_MAX/2,10},43,69,0, {0,0,0,0},0,{0,0,0,0,0,0,0,0}},
  { "V1 Squ",  0,0,0,0,{0,0,127, 0,0,0,   0},{20,10,ENV_MAX/2,10},43,69,0, {0,0,0,0},0,{0,0,0,0,0,0,0,0}},
  { "V1 Tri",  0,0,0,0,{127,0,0, 0,0,0,   0},{20,10,ENV_MAX/2,10},43,69,0, {0,0,0,0},0,{0,0,0,0,0,0,0,0}},
  { "V2 Saw",  0,0,0,0,{0,0,0,   0,127,0, 0},{20,10,ENV_MAX/2,10},43,69,0, {0,0,0,0},0,{0,0,0,0,0,0,0,0}},   
  { "V2 Squ",  0,0,0,0,{0,0,0,   0,0,127, 0},{20,10,ENV_MAX/2,10},43,69,0, {0,0,0,0},0,{0,0,0,0,0,0,0,0}},
  { "V2 Tri",  0,0,0,0,{0,0,0,   127,0,0, 0},{20,10,ENV_MAX/2,10},43,69,0, {0,0,0,0},0,{0,0,0,0,0,0,0,0}},
  { "2x All",  0,0,0,0,{64,64,64,64,64,64,0},{20,10,ENV_MAX/2,10},43,69,0, {0,0,0,0},0,{0,0,0,0,0,0,0,0}},
  { "NoWave",  0,0,0,0,{0,0,0,   0,0,0,   0},{20,10,ENV_MAX/2,10},43,127,0,{0,0,0,0},0,{0,0,0,0,0,0,0,0}},
};

const int16_t WaveSine[257] = {
     0,   804,  1608,  2410,  3212,  4011,  4808,  5602,  6393,  7179,
  7962,  8739,  9512, 10278, 11039, 11793, 12539, 13279, 14010, 14732,
 15446, 16151, 16846, 17530, 18204, 18868, 19519, 20159, 20787, 21403,
 22005, 22594, 23170, 23731, 24279, 24811, 25329, 25832, 26319, 26790,
 27245, 27683, 28105, 28510, 28898, 29268, 29621, 29956, 30273, 30571,
 30852, 31113, 31356, 31580, 31785, 31971, 32137, 32285, 32412, 32521,
 32609, 32678, 32728, 32757, 32767, 32757, 32728, 32678, 32609, 32521,
 32412, 32285, 32137, 31971, 31785, 31580, 31356, 31113, 30852, 30571,
 30273, 29956, 29621, 29268, 28898, 28510, 28105, 27683, 27245, 26790,
 26319, 25832, 25329, 24811, 24279, 23731, 23170, 22594, 22005, 21403,
 20787, 20159, 19519, 18868, 18204, 17530, 16846, 16151, 15446, 14732,
 14010, 13279, 12539, 11793, 11039, 10278,  9512,  8739,  7962,  7179,
  6393,  5602,  4808,  4011,  3212,  2410,  1608,   804,     0,  -804,
 -1608, -2410, -3212, -4011, -4808, -5602, -6393, -7179, -7962, -8739,
 -9512,-10278,-11039,-11793,-12539,-13279,-14010,-14732,-15446,-16151,
-16846,-17530,-18204,-18868,-19519,-20159,-20787,-21403,-22005,-22594,
-23170,-23731,-24279,-24811,-25329,-25832,-26319,-26790,-27245,-27683,
-28105,-28510,-28898,-29268,-29621,-29956,-30273,-30571,-30852,-31113,
-31356,-31580,-31785,-31971,-32137,-32285,-32412,-32521,-32609,-32678,
-32728,-32757,-32767,-32757,-32728,-32678,-32609,-32521,-32412,-32285,
-32137,-31971,-31785,-31580,-31356,-31113,-30852,-30571,-30273,-29956,
-29621,-29268,-28898,-28510,-28105,-27683,-27245,-26790,-26319,-25832,
-25329,-24811,-24279,-23731,-23170,-22594,-22005,-21403,-20787,-20159,
-19519,-18868,-18204,-17530,-16846,-16151,-15446,-14732,-14010,-13279,
-12539,-11793,-11039,-10278, -9512, -8739, -7962, -7179, -6393, -5602,
 -4808, -4011, -3212, -2410, -1608,  -804,     0
};

/**** OLED Functions ***/
void u8g2_prepare(void) {
  u8g2.setFont(u8g2_font_6x10_tf);
  u8g2.setFontRefHeightExtendedText();
  u8g2.setDrawColor(1);
  u8g2.setFontPosTop();
  u8g2.setFontDirection(0);
}

void PolyDisp(void) {
//  char buf1[16];
int barh, i;
#define BARMAX  16
#define BARX0 8

  u8g2.clearBuffer();
  u8g2_prepare();
  u8g2.setCursor( 0, 0);
  u8g2.print( "Patch ");
  u8g2.print(patchSel);
  u8g2.print(" ");
  u8g2.print(patch[patchSel].Pname);
  u8g2.drawStr( 0, 24, "TR SA SQ TR SA SQ NO");
  for (i=0; i< 7; i++){
    barh = patch[patchSel].Mix[i] / 8 ;
 //   barh  = 2 * i;
    u8g2.drawBox(i*18, BARX0 + BARMAX - barh, 12, barh);
  }
  
  u8g2.setCursor( 0, 48);
  u8g2.print("nextvoice ");
  u8g2.print(nextvoice);
  
  u8g2.setCursor( 0, 56);
  u8g2.print("Note ");
  u8g2.print(vco[1].newnote);
  
  u8g2.setCursor( 46, 56);
  u8g2.print("V2C ");
  u8g2.print(coarseDetuneV2 * 12 / 500);

  u8g2.setCursor( 92, 56);
  u8g2.print("V2F");
  u8g2.print(fineDetuneV2);
  
  u8g2.sendBuffer();
}


void updateEnv(void){
  int v = 0;        // Need a loop here
  
  for(v = 0; v< voices; v++){
    switch (amp_env.state[v]) {
    case ENV_IDL:      // Idle state, Out = 0 
      amp_env.out[v] = 0;
      break;
    case ENV_ATT:      // Attack state, linear ramp
      amp_env.out[v] += amp_env.att;
      if(amp_env.out[v] > ENV_MAX){
        amp_env.out[v] = ENV_MAX;
        amp_env.state[v] = ENV_DEC;
      }
      break;
    case ENV_DEC:      // Decay state, linear ramp until SUS reached, then hold
      amp_env.out[v] -= amp_env.dec;
      if(amp_env.out[v] < amp_env.sus){
        amp_env.out[v] = amp_env.sus;
      }
      break;
    case ENV_REL:      // Release state, linear ramp towards 0. Should be exp. decay
      amp_env.out[v] -= amp_env.rel;
      if(amp_env.out[v] < 0){
        amp_env.out[v] = 0;
        amp_env.state[v] = ENV_IDL;
      }   
      break;
    default:
      amp_env.state[v] = ENV_IDL;
      break;
    } 
  
    amp_env.dac[v] = amp_env.out[v] * 4095 / ENV_MAX;   // Scale Envelope to DAC
  }
}

void updateLfo(void){
//  lfo1.freq = 0.2;
  lfo1.ph_inc  = int(4294967296.0 * lfo1.freq / F_SAMPLE);    // Do this outside the update()
  lfo1.ph_acc += lfo1.ph_inc;
//  lfo1.out = lfo1.ph_acc >> 20;     // 32 bits - 12 bits = 20
  lfo1.out = (lfo1.ph_acc >> 24) & 0x00ff;
  lfo1.out = (WaveSine[lfo1.out]>>4) + 2048; //table is 16 bits signed, make it unsigned 12 bits
  
}

// Use for setting board and chip adresses. 
void BAddr(char ba){
  ba &= 7;
  digitalWrite(BA0, ba & 1);
  digitalWrite(BA1, ba & 2);
  digitalWrite(BA2, ba & 4);
}
void CAddr(char ca){
  ca &= 7;
  digitalWrite(CA0, ca & 1);
  digitalWrite(CA1, ca & 2);
  digitalWrite(CA2, ca & 4);
}


// OLD CODE
/*
void DAC4822Wr(int ch, int val)  // Fast Bit Bang, 7MHz clock Using ditialWrite() was 1MHz
{
  if (ch == 0) val = ((val & 0x0FFF) | 0x7000);  // MSB is 0 
  else val = ((val & 0x0FFF) | 0xF000);          // MSB is 1 
  GPIOD_PCOR = 1 << SPI_SCK;                     // SCK = 0
  GPIOD_PCOR = 1 << SPI_SS12;                    // SS = 0
  for (int i=0; i< 16; i++){
    if (val & 0x8000) GPIOD_PSOR = 1 << SPI_MOSI;// MOSI = 1
    else GPIOD_PCOR = 1 << SPI_MOSI;             // MOSI = 0
    GPIOD_PSOR = 1 << SPI_SCK;                   // SCK = 1 
    GPIOD_PCOR = 1 << SPI_SCK;                   // SCK = 0                   
    val = val << 1;
  }
  GPIOD_PSOR = 1 << SPI_SS12;                    // SS = 1 
} */

// Fast bit bang for SPI Teensy 3.5
#define C_SS  GPIOB_PCOR      // Clr Bit Port B11
#define S_SS  GPIOB_PSOR      // Set Bit Port B11
#define B_SS   10
#define C_MOSI GPIOD_PCOR     // Clr Bit Port B11
#define S_MOSI GPIOD_PSOR     // Set Bit Port B11
#define B_MOSI 4
#define C_SCK  GPIOB_PCOR     // Clr Bit Port B10
#define S_SCK  GPIOB_PSOR     // Set Bit Port B10
#define B_SCK  11  
/* Low level 16 bit SPI Write using ports 
 *  pol is clock polarity: 0 = CK starts LO bits clocked on H-L, (DAC108)
 *                         1 = CK starts HI bits clocked on L-H (MCP4822) */
void SPIWr(char pol, int val){                  
  S_SS = 1 << B_SS;                            // SS = 1
  if (pol == 0){               
    C_SCK = 1 << B_SCK;                        // CK = 0
    C_SS = 1 << B_SS;                          // SS = 0 
    for (int i=0; i< 16; i++){    
      if (val & 0x8000) S_MOSI = 1 << B_MOSI;  // MOSI = 1
      else C_MOSI = 1 << B_MOSI;               // MOSI = 0              
      S_SCK = 1 << B_SCK;                      // Set Clock 
      C_SCK = 1 << B_SCK;                      // Reset Clock
      val = val << 1;     
    }               
  }
  else{
    S_SCK = 1 << B_SCK; 
    C_SS = 1 << B_SS;                           // SS = 0    
    for (int i=0; i< 16; i++){
      if (val & 0x8000) S_MOSI = 1 << B_MOSI;   
      // MOSI = 1
      else C_MOSI = 1 << B_MOSI;                // MOSI = 0
      C_SCK = 1 << B_SCK;                       // Reset Clock 
      S_SCK = 1 << B_SCK;                       // Set Clock
      val = val << 1;
    } 
  }
  S_SS = 1 << B_SS;                             // SS = 1  
}
/* Low level 16 bit SPI Write using digitalWrite() 
 *  pol is clock polarity: 0 = CK starts LO bits clocked on H-L, (DAC108)
 *                         1 = CK starts HI bits clocked on L-H (MCP4822) */
/*void SPIWr(char pol, int val){
  digitalWrite(SPI_SS , 1);                    // SS = 1

  if (pol == 0){
    digitalWrite(SPI_CK , 0);                    // CK = 0
    digitalWrite(SPI_SS , 0);                    // SS = 0                                  
    for (int i=0; i< 16; i++){    
      if (val & 0x8000) digitalWrite(SPI_MOSI, 1);  // MOSI = 1
      else digitalWrite(SPI_MOSI, 0);               // MOSI = 0              
      digitalWrite(SPI_CK, 1);                      // Set Clock 
      digitalWrite(SPI_CK, 0);                      // Reset Clock
      val = val << 1;     
    }               
  }
  else{
    digitalWrite(SPI_CK , 1); 
    digitalWrite(SPI_SS , 0);                    // SS = 0   
    for (int i=0; i< 16; i++){
      if (val & 0x8000) digitalWrite(SPI_MOSI, 1);// MOSI = 1
      else digitalWrite(SPI_MOSI, 0);             // MOSI = 0
      digitalWrite(SPI_CK, 0);                      // Reset Clock 
      digitalWrite(SPI_CK, 1);                      // Set Clock
      val = val << 1;
    } 
  }
  digitalWrite(SPI_SS , 1);                      // SS = 1  
}
*/

void DAC12Wr( char ch, int val){
  if (ch == 0) val = ((val & 0x0FFF) | 0x7000);  // MSB is 0 
  else val = ((val & 0x0FFF) | 0xF000);          // MSB is 1 
  CAddr(0);
  SPIWr(0, val);
}

void DAC108Wr(int ch, int val){  // 10b DAC takes 12b value, ignores 2 LSBs
  // Note: DAC108 clock requires negative polarity clock. 
  // 4 MSBs are dac set if 0-7, command if 8-F
  val = (val & 0x0FFC) | ((ch & 0x000f) << 12);  // channel is 3 bits [14:12], MSB is 0 
  CAddr(1);
  SPIWr(1, val);
}

void digPot(int ch, int val){   // 8 channels, Ch 0-3 CA3, 4-7 CA2 ? 4 data 00 val 0-256 (100h) 
  ch &= 0x07;                   // Clip the parameters
  if (val > 256) val = 256;     
  if (ch > 3) CAddr(2);         // Pick a chip address: 0-3 CA3, 4-7 CA2
  else CAddr(3);                 
  ch &= 0x03;                   // 2 bit channel address within the chip
  if (ch & 0x2) ch += 4;        // 2,3 -> 6,7
  val = ((ch & 0x0F) << 12) | val;  // channel is 4 bits [14:12],  write is 00, data is 10 bits, 0-256
  SPIWr(0, val);
}

// This needs to be fast to update 8 voices. BitBang OK for 1 channel
void setDacs(){  
  int v;

  updateEnv();        //by putting these here, can measure update time on scope. 
  updateLfo();
  
//  BAddr(0);
//  CAddr(0);

// Set Pitch DACs
 for (v=0; v<voices; v++){
    BAddr(v);
    DAC12Wr(0, vco[v].V1dac);
    DAC12Wr(1, vco[v].V2dac);
  }

  // DAC108 is 5V so 4.88mV/LSB 10b, 1.22mV 12b
  for (v=0; v<voices; v++){
    BAddr(v);
    DAC108Wr(9, 0);       // Set to WTM Write Thru Mode
    DAC108Wr(D_MOD3, 2048);
    DAC108Wr(D_MOD4, 2048);
    DAC108Wr(D_PW1, 2048);
    DAC108Wr(D_PW2, 2048);
    DAC108Wr(D_SPARE, lfo1.out);    // Put out test patterns on test point
                                    // Need a loop here
    DAC108Wr(D_FILT_F, fil.freq);     
    DAC108Wr(D_FILT_R, fil.res);     
    DAC108Wr(D_VCA, amp_env.dac[v]);
  }
 
  if (mix.flag){       // Update mixer
    for (v=0; v<voices; v++){
      BAddr(v);
   // Serial.println("Up Mix");
      digPot(M_V1_TRI, patch[patchSel].Mix[V1TRI]);
      digPot(M_V1_SAW, patch[patchSel].Mix[V1SAW]);
      digPot(M_V1_SQU, patch[patchSel].Mix[V1SQU]);
      digPot(M_V2_TRI, patch[patchSel].Mix[V2TRI]);
      digPot(M_V2_SAW, patch[patchSel].Mix[V2SAW]);
      digPot(M_V2_SQU, patch[patchSel].Mix[V2SQU]);
      digPot(M_NOISE, 0);
    }
    mix.flag = 0;
  }
}  // End of setDacs()

// Sets Note, Pitch wheel, cal, and detune for pitch DACs
void setPitch(int v){
  int V1Off, V2Off;   // Calibration offsets. Should be in setup()
  int board;
  board = v + 1; // For now, voices 0-7, boards 1-3
  V1Off = (int) (boardCal[board].v1Offs * 500.0 / 12.0 + 0.5) ;  // Set freq offset cal here in semitones. 500/12 counts per semitone
  V2Off = (int) (boardCal[board].v2Offs * 500.0 / 12.0 + 0.5); 
  vco[v].V1dac = vco[v].newnote * 500 / 12 + vco[v].pitchDac + V1Off;  
  vco[v].V2dac = vco[v].newnote * 500 / 12 + vco[v].pitchDac + V2Off + fineDetuneV2 + coarseDetuneV2;  
  if(vco[v].V1dac > 4095) vco[v].V1dac = 4095;        // And clip
  if(vco[v].V2dac > 4095) vco[v].V2dac = 4095;
}
                                                               
void patchToHW (void){
  mix.flag = 1;        // Update mixer
}


/****    MIDI Callbacks ****/

void OnNoteOn(byte channel, byte note, byte velocity) {
  int v;
  if(unison){
    for(v=0; v<voices; v++){
      vco[v].newnote = note;        // Just Unison for now
      setPitch(v);
      amp_env.state[v] = ENV_ATT;   // Start the envelope
    }
  }
  else{
    vco[nextvoice].newnote = note;        // Just count thru voices for now
    setPitch(nextvoice);
    amp_env.state[nextvoice] = ENV_ATT;   // Start the envelope     
    Serial.print("On ");
    Serial.print(note);
    Serial.print(" voice ");
    Serial.println(nextvoice);
    nextvoice++;
    if (nextvoice > (voices - 1)) nextvoice = 0; 
  }
}

void OnNoteOff(byte channel, byte note, byte velocity) {
  int v;
  if (unison){
    for (v=0; v<voices; v++)        // For unison: turn off all
     amp_env.state[v] = ENV_REL;
  }
  else{
    for (v=0; v<voices; v++) {      // Find note
      if (vco[v].newnote == note){
        amp_env.state[v] = ENV_REL;
        Serial.print("Off ");
        Serial.print(note);
        Serial.print(" Voice ");
        Serial.println(v);
        //return;
      }
    }
  }
}

// Pitch bend range is +8191 to -8192, typical resolution is 128 counts
void OnPitchBend(byte channel, int bend){
  int v;
  pitchDac = 83 * bend / 8192;     // 83 dac codes is 2 semitones 
// 500 dac codes per octave (12 semitones)
  for (v=0; v<voices; v++)        // gotta tweak all voices. 
    setPitch(v);
//  Serial.print("Bend: ");
//  Serial.println(bend);
}

void OnControlChange(byte channel, byte control, byte value) {
  int v;
// Serial.print("Ctl: ");
//  Serial.print(control);
//  Serial.print(" Val: ");
//  Serial.println(value);
  if (control == 07) {             // Volume control change message

  }
  else {
    switch (control) {
    case 0x40:                // damper/sustain pedal
      break;
    case OXC1:                // Oxygen 49 drawbars: C1-C9
      fil.freq = value * 4095 / 127; 
      break;
    case OXC2:
      fil.res = value * 4095 / 127;
      break;
    case OXC3:
      break;
    case OXC4:
      break;
    case OXC5:
      break;
    case OXC6:
      break;
    case OXC7:
      break;
    case OXC8:
      break;
    case OXC9:
     lfo1.freq = (value+1) / 5.0;  //128 = 25.6 // +1 for no 0.0 Hz
      break;
    case OXC10:     // Oxygen Knobs: C10-C17  C10 coarse pitch detune 
      coarseDetuneV2 = 500 / 12 * ((value-64)/4);      // 16 semitones - 1.3 octaves 
      for (v=0; v<voices; v++)        // gotta tweak all voices. 
        setPitch(v);
      break;
    case OXC11:     // C11 fine pitch detune 
      fineDetuneV2 = (value-64)/2;
      for (v=0; v<voices; v++)        // gotta tweak all voices. 
        setPitch(v);
      break;
    }
  }    
}

// Program change buttons OXY49 C18-C26
void OnProgramChange(byte channel, byte program ) {
  Serial.print("Chan:");
  Serial.print(channel);
  Serial.print(" Progl: ");
  Serial.println(program);
  switch (program) {
    case 0:                // Oxy49 Button C18 toggles thru patches
      patchSel++;
      if (patchSel > (NUMPATCH -1)) patchSel = 0;
      Serial.print("Patch ");
      Serial.print(patchSel);
      Serial.print(" ");
      Serial.println(patch[patchSel].Pname);
      patchToHW();
      break;
    case 1:               // Oxy49 Button C19       
      break;
    case 2:               // Oxy49 Button C20 
      break;
  }
}


void setup() {
  
  Serial.begin(115200);
  delay(1000); // delay to allow USB serial to come up
  Serial.println("DIY Poly on Teensy 3.5 E.E.");
  
  MIDI.begin(MIDI_CHANNEL_OMNI);
//  MIDI.begin();
  MIDI.setHandleNoteOn(OnNoteOn);  
  MIDI.setHandleNoteOff(OnNoteOff);  
  MIDI.setHandleControlChange(OnControlChange);   
  MIDI.setHandleProgramChange(OnProgramChange); 
  MIDI.setHandlePitchBend(OnPitchBend);
//  MIDI.sendProgramChange(instrument, channel);

// Initialize SPI pins
  pinMode(BA0, OUTPUT);
  pinMode(BA1, OUTPUT);
  pinMode(BA2, OUTPUT);
  pinMode(CA0, OUTPUT);
  pinMode(CA1, OUTPUT);
  pinMode(CA2, OUTPUT);
  pinMode(SPI_SS, OUTPUT);
  digitalWrite(SPI_SS, 1);
  pinMode(SPI_CK, OUTPUT);
  digitalWrite(SPI_CK, 0);
  pinMode(SPI_MOSI, OUTPUT);
  digitalWrite(SPI_MOSI, 0);

  u8g2.begin();
 
  Timer1.initialize(1000000 / F_SAMPLE); // in uS
  Timer1.attachInterrupt(setDacs); // Update DACs on timer 1 interrupt

  pitchDac = 0;

  int v;
  for (v=0; v<voices; v++){      // Initialize vco params  
    vco[v].newnote = 0;
    vco[v].order = 0;
  } 
  // Initialize Env params. For now low is slow. 50 is about 
  amp_env.att = 50;
  amp_env.dec = 100;
  amp_env.sus = ENV_MAX / 2;
  amp_env.rel = 20;
  for (v=0; v < voices; v++){
    amp_env.state[v] = ENV_IDL;
    amp_env.out[v] = 0;
  }
  // Initialize Filters
  fil.freq = 4095;    // Max BW
  fil.res  = 1024;

  Serial.println("End of Setup()");
}

void loop() {
//  Serial.println("LOOP!");
  MIDI.read();    // MIDI receive, slowed by processing in loop() (OLED). Should be in a fast interrupt
  
  PolyDisp();

//  delay (5);  // This causes MIDI slowdown.  OLED update does also. 
}







